# -*- coding: utf-8 -*-
"""A clock based on LSL local_clock
"""
from pylsl import local_clock


class Clock():
    """allows to keep track of time

    reset the clock with :meth:`~.reset`, and get the time since this reset with :meth:`~.now`. Measure intermediate times with :meth:`~.tick`, which return time since last call of :meth:`tick`, or

    each call of :meth:`~.tick` updates the cumulative time since the last call of :meth:`reset`. If you don't want to add the time measured, use :meth:`~.pause`
    """

    def __init__(self):
        "create a new instance requiring no arguments"
        self._error = 0.
        self.reset()

    def time(self):
        "time in seconds (usually since computer boot) according to LSL"
        return local_clock()

    def tick(self) -> float:
        """time since last call of :meth:`~.tick`

        returns
        -------
        delta_t: float
            time passed in seconds since the last call of :meth:`~.tick` or
            :meth:`~.pause`. This is time counting into the cumulative time


        """
        ts = self.time()
        delta_t = ts - self.last_tick
        self.last_tick = ts
        return delta_t

    def sleep(self, duration: float):
        """Sleep for duration in seconds

        blocking is slightly more accurate than non-blocking sleep, e.g. as available with :meth:`time.sleep` from the stdlib. There is one disadvantage of this kind of busy sleep: it can cause slight oversleeping, as there is an overhead of the function being called and returning, which is not accounted for. See :meth:`~.sleep_debiased` for an alternative sleep with asymptotic minimisation of the error.

        args
        ----
        duration: float
            how many seconds to sleep blocking

        returns
        -------
        duration: float
            the time in seconds spent sleeping
        """
        t1 = t0 = self.time()
        dt = 0
        while dt <= duration:
            t1 = self.time()
            dt = t1-t0
        return dt

    def sleep_debiased(self, duration: float):
        """Sleep for duration in seconds with attempts for debiasing`

        sometimes, you execute some other commands, and these commands have a variable runtime. If we would naively sleep everytime for n seconds afterwards, we would inherit this jitter. By using :meth:`~.tick` before these commands, and :meth:`~.sleep_debiased` after these commands, we can normalize the runtime to a fixed period (as long as the sleep duration is longer than the runtime of the commands).

        Additionally, this function keeps track of any oversleeping or undersleeping, and will minimize the temporal error asymptotically

        returns
        -------
        duration: float
            the time in seconds spent sleeping since the last call
            of :meth:`~.tick` or :meth:`~.sleep_debiased`.


        Example
        -------

        This example shows how we can regularize the time spent in each cycle to 200ms in spite of there being an element of random runtime

        .. code-block:: python

            import time
            import random
            from reiz.time import Clock
            clock = Clock()
            t = 0.
            msg = "{0:3.5f}, {1:3.5f}, {2:3.5f}, slept for {3:3.5f}s"
            for i in range(1, 11):
                time.sleep(random.random()/10)
                dt = clock.sleep_debiased(0.2)
                t += dt
                print(msg.format(i*0.2, clock.now(), t, dt))
        """
        bias = self.tick()
        dt = self.sleep(duration - bias - self._error)
        self._error += dt + bias - duration
        tick_bias = self.tick()-dt
        self._error += tick_bias
        return dt + bias + tick_bias

    def reset(self):
        """reset the clock

        resets the counter keeping track of the cumulative time spend since instantiaion or the last call of :meth:`~.reset`
        """
        self._t0 = self.last_tick = self.time()

    def now(self):
        """return the cumulative time passed since the last call of :meth:`reset`
        """
        return self.time()-self._t0


clock = Clock()  #: a default :class:`.Clock` instance ready for your experiment



# -*- coding: utf-8 -*-
"""
Building blocks for auditory stimuli
....................................
"""

from .tts import tts_Mixin
from sys import platform
from threading import Timer
import time
from pyglet.media.codecs.base import StaticSource
from pyglet.media.synthesis import ADSREnvelope, Sine
import pyglet

# %%


class Sound():
    """interface for a generic sound object"""
    source = None
    volume = 1

    def play(self):
        """start playing the sound and return immediatly

        returns
        -------
        remainder: float
            seconds until the sound would be finished playing
        """
        t0 = time.time()
        player = self.source.play()
        player.volume = self.volume
        t = Timer(player.source.duration, player.delete)
        t.start()
        return player.source.duration + t0 - time.time()

    def play_blocking(self) -> float:
        """"play the sound and block until playing is finished

        returns
        -------
        remainder: float
            seconds until the sound would finish playing. Naturally, it always
            returns 0. The output is kept here to allow easy substitution with
            : func: `Sound.play`
        """
        rest = self.play()
        t0 = time.time()
        while (time.time()-t0) < rest:
            pass
        return 0

    @property
    def duration(self) -> float:
        "returns duration of the sound in seconds"
        return self.source.duration

    def __repr__(self):
        return str(self.source)


class AnySource(Sound):
    """
    args
    ----
    source: StaticSource
        the source of the sound. Usually you instantiate one of the
        inheriting objects, e.g. :class:`~.AudioFile` or :class`~.Hertz`
        which allow more parameters.

    """

    def __init__(self, source: StaticSource = None):
        self.source = source


class Hertz(Sound):
    """instantiate a sine wave with ramp-up and ramp-down period of volume

    args
    ----
    duration_in_ms: float
        duration of the sound in ms
    frequency: int
        frequency in Hz
    volume: float
        relative volume
    """

    def __init__(self, duration_in_ms: float = 500, frequency: int = 440, volume: float = 1):

        adsr = ADSREnvelope(attack=duration_in_ms/8000,
                            decay=duration_in_ms/2000,
                            release=duration_in_ms/8000,
                            sustain_amplitude=volume)
        sine = Sine(duration=duration_in_ms/1000, frequency=frequency,
                    sample_size=16, sample_rate=44100, envelope=adsr)
        self._duration_in_ms = duration_in_ms
        self._frequency = frequency
        self.source = pyglet.media.StaticSource(sine)
        self.volume = volume

    def __repr__(self):
        return f"Hertz(duration_in_ms={self._duration_in_ms}, frequency= {self._frequency})"


class AudioFile(Sound):
    """instantiate a sound object from a wav-file

    args
    ----
    filepath: str
        path to the wav-file
    """

    def __init__(self, filepath: str):
        self.source = pyglet.media.load(filepath, streaming=False)


class Message(tts_Mixin, Sound):
    """Text-to-Speech Audio Stimulus

    If reiz was installed with `[tts]` as extra, this class will use pyttsx3
    to generate auditory stimuli on demand. If pyttsx3 is not found, there will be only silence when creating an `audio.Message`.
    """
    pass



# -*- coding: utf-8 -*-
"""
Created on Mon May 20 10:50:52 2019

@author: Robert Guggenberger
"""
from tempfile import NamedTemporaryFile
from subprocess import run
from typing import Callable
from threading import Thread
import time
import pyglet.media
from sys import platform
import importlib


class PlatformIndependentMessage():
    """instantiate a tts sound object

    args
    ----
    message: str
        the content of the message, e.g. "Hello World"
    voiceid:int
        the id of the voice
    rate:int
        the speed of the Utterance
    """

    def __init__(self, message: str = "Hello World",
                 rate=135, voiceid=0):
        if "darwin" in platform:
            self.engine = pyttsx3.init("dummy")
        else:
            self.engine = pyttsx3.init()
        self.engine.setProperty('rate', rate)
        voices = self.engine.getProperty('voices')
        self.engine.setProperty('voice', voices[voiceid])
        self.message = message
        self._duration = self._estimate_duration()

    def _estimate_duration(self):
        t0 = time.time()
        self.engine.setProperty('volume', 0.)
        self.play_blocking()
        t1 = time.time()
        self.engine.setProperty('volume', 1.)
        return t1-t0

    @property
    def volume(self):
        return self.engine.getProperty('volume')

    @volume.setter
    def volume(self, volume: float):
        self.engine.setProperty('volume', volume)

    def play_blocking(self):
        """
        returns
        -------
        remainder:float
            seconds until the sound would finish playing. Naturally, it always
            returns 0. The output is kept here to allow easy substitution with
            :func:`Sound.play`
        """
        self.engine.say(self.message)
        self.engine.runAndWait()
        return 0

    def stop(self):
        self.engine.stop()

    def play(self):
        """start playing the sound and return immediatly

        returns
        -------
        remainder:float
            seconds until the sound would be finished playing
        """
        t0 = time.time()
        t = Thread(target=self.play_blocking)
        t.start()
        return self.duration + t0 - time.time()

    @property
    def duration(self) -> float:
        "returns duration of the sound in seconds"
        return self._duration

    def __repr__(self):
        return f"{self.__class__} {str(self.message)}"


class Espeak_Mixin(object):

    def __init__(self, message: str = "Espeak says",
                 rate: int = 135,
                 gender="f",
                 language="de"):
        self.message = message
        if gender.lower() not in ["m", "f"]:
            raise ValueError("Gender must be (m)ale or (f)emale")

        with NamedTemporaryFile(suffix=".wav") as f:
            run(["espeak", f"\"{message}\""
                 "-s", str(rate),
                 f"-v{language.lower()}+{gender.lower()}1",
                 # increase  pitch for words which begin a capital letter.
                 "-k10",
                 "-g2",
                 "-w", f.name])
            self.source = pyglet.media.load(f.name, streaming=False)


class Silent_Mixin(object):
    def __init__(self, message: str = "Silence is golden",
                 language="de"):
        self.message = message
        print("no TTS is installed on your system. Default to silence")
        from pyglet.media.synthesis import Silence
        self.source = Silence(duration=0.1)
        return


# class gTTS_Mixin():

#     def __init__(self, message: str = "Google says",
#                  language="de"):
#         self.message = message
#         try:
#             from gtts import gTTS
#         except ImportError:
#             print("gTTS is not installed on your system. Default to silence")
#             from pyglet.media.synthesis import Silence
#             self.source = Silence(duration=0.1)
#             return

#         with NamedTemporaryFile(suffix=".wav") as f:
#             tts = gTTS(message, lang=language)
#             tts.save(f.name)
#             run(["ffmpeg", "-i", f.name, f.name, "-y"])
#             self.source = pyglet.media.load(f.name, streaming=False)


# conditional interface for Message

if importlib.util.find_spec("pyttsx3") is None:  # pragma: no cover
    tts_Mixin = Silent_Mixin
else:
    if "darwin" in platform:  # pragma: no cover
        tts_Mixin = Silent_Mixin
    else:
        import pyttsx3
        tts_Mixin = PlatformIndependentMessage



"""Sending markers

Sending markers requires a MarkerServer running. This markerserver is used as a man-in-the-middle to forward all markers into an already existing LSLOutlet.
This allows easier recording, reduces computational load on your main script, and prevents forgetting to record a MarkerStream with LabRecorder. It adds a little bit more complexity and a miniscule delay.

You can safeguard your script to make sure a MarkerServer is running in at least these two ways:

a) check whether the MarkerServer is avaible, and abort with a notice, if that is not the case.
b) start a throw-away MarkerServer with :meth:`~.start`. 

Sending markers
---------------

There is not much need to push markers for stimuli events manually, as you can create a :class:`~.Cue` and add the desired marker-string during instantiation. Often, you want to send additional information, e.g. the age and id of the subject, or the parameters with which the experiment was run. In these cases, pushing dictionaries or arbitrary strings during initialization of your experiment can be useful.  

.. currentmodule:: reiz._marker.client
.. autosummary::
   :template: module.rst

    push
    push_json



Safeguarding
------------
.. currentmodule:: reiz._marker.safeguard
.. autosummary::
   :template: module.rst

    available
    start
    stop

"""

from reiz._marker.client import push, push_json, available
from reiz._marker.safeguard import start, stop



import reiz


def test_visual_library():
    "contains Image, Cross, Mural and Trapezoid"
    c = reiz.Canvas()
    for item in reiz.visual.library.__dict__.values():
        item.draw(c)


def test_colors():
    "also draws a background"
    from reiz._visual.colors import COLORS
    from reiz._visual.colors import resolve_rgb
    canvas = reiz.Canvas()
    for c in COLORS:
        resolve_rgb("white", c)
        reiz.visual.Background(color=c).draw(canvas)


def test_smoke_complex():
    canvas = reiz.Canvas()
    reiz._visual.complex.Line().draw(canvas)
    reiz._visual.complex.Polygon().draw(canvas)
    reiz._visual.complex.Bar().draw(canvas)
    reiz._visual.complex.Circle().draw(canvas)
    reiz._visual.complex.Cylinder().draw(canvas)
    reiz._visual.complex.Mural().draw(canvas)
    reiz._visual.complex.Trapezoid().draw(canvas)
    reiz._visual.complex.Cross().draw(canvas)


def test_mural():
    canvas = reiz.Canvas()
    m = reiz._visual.complex.Mural(text="test")
    assert repr(m) == "Mural('test')"


def test_cross():
    canvas = reiz.Canvas()
    m = reiz._visual.complex.Cross(zoom=1, color=(1, 0, 0))
    assert repr(m) == "Cross(zoom=1, color=(1, 0, 0))"


def test_iteration():
    canvas = reiz.Canvas()
    for v in reiz._visual.complex.Line():
        v.draw(canvas)

    for v in [reiz._visual.complex.Line(), reiz._visual.complex.Line()]:
        v.draw(canvas)



"""Creating and managing visual stimuli
"""
import os
from os import environ as _env
from reiz._visual.complex import Mural, Image, Cross, Circle, Background, Line, Bar
from reiz._visual.complex import Polygon, Trapezoid, Cylinder
from reiz._visual.colors import COLORS, get_color
from types import SimpleNamespace
from typing import Dict, NewType, Any
from pathlib import Path
from pkg_resources import resource_filename

libConf = NewType("libConf", Dict[str, Dict[str, Any]]
                  )  #: Dict[str, Dict[str, Any], a dictionary of types and respective keywords arguments

_defaults = libConf({
    "Mural": {
        "post": {"text": "Run endet"},
        "pre": {"text": "Run beginnt"},
        "eo": {"text": "Augen offen"},
        "ec": {"text": "Augen zu"},
        "ready": {"text": "Bereit machen"},
        "los": {"text": "Los"},
        "go": {"text": "Go!"},
        "imagine": {"text": "Bewegung vorstellen"},
        "move": {"text": "Bewegung starten"},
        "count": {"text": "Zahlen berechnen"},
        "relax": {"text": "Entspannen"},
        "rating": {"text": "0 - 1 - 2 - 3 - 4 - 5 - 6 - 7 - 8 - 9 - 10", "fontsize": 0.75},
    },
    "Cross": {
        "fixation": {"color": "white"},
    },

    "Image": {
        "logo": {"imgpath": resource_filename(__name__, 'data/logo.png')},
    }

})  #: libConf


def make_library(settings: libConf = None, failraise=False) -> SimpleNamespace:
    """create a library of visual stimuli from a dictionary of arguments

    args
    ----
    settings:
        a dictionary of types with kwargs appropriate to the respective type
        as dictionary
    failraise:bool
        raise an exception if a key is not recognized. defaults to False

    returns
    -------

    library:
        a library of instances of stimuli which can be adressed in dot.notation
    """

    if settings is None:
        settings = _defaults
    library = dict()
    for key in settings.keys():
        for name, args in settings[key].items():
            if key.lower() == "mural":
                library[name] = Mural(**args)
            elif key.lower() == "cross":
                library[name] = Cross(**args)
            elif key.lower() == "image":
                library[name] = Image(**args)
            else:
                if failraise:
                    raise(f"{key} with {args} can't be processed")

    library = SimpleNamespace(**library)
    return library


if not 'DOC' in _env.keys():
    library = make_library()
else:   # pragma: no cover
    mock = dict()
    for t, item in _defaults.items():
        for k, v in item.items():
            mock[k] = t
    library = SimpleNamespace(**mock)  #: a library of visual stimuli
    print("Generating sphinx documentation. Skipping visual library")


__doc__ += f"""
During import, a basic library will be created. By default, it contains
the following set of visual stimuli: {', '.join([str(k) for k in library.__dict__])}. These stimuli can be addressed by dot.notation from :data:`reiz.visual.library`, for example like :data:`reiz.visual.library.go`

.. automodule:: reiz._visual.complex
   :members: Mural, Image, Cross, Circle, Background, Line, Bar, Polygon, Trapezoid, Cylinder

.. automodule:: reiz._visual.colors
   :members: COLORS, get_color

Create libraries of visual stimuli
..................................



"""



# -*- coding: utf-8 -*-
"""
Pyglet based screen and frame drawing

    
"""
import pyglet
from pylsl import local_clock
from typing import Tuple
# %%


def get_screens():
    return pyglet.canvas.Display().get_screens()

# %%


class ExperimentalWindow(pyglet.window.Window):
    start_run = False
    paused = False

    def on_close(self):
        """Default on_close handler, but closing even without vent loop"""
        self.has_exit = True
        #from pyglet import app
        # if app.event_loop.is_running:
        self.close()

    def on_key_press(self, symbol, modifiers):
        """Default on_key_press handler."""
        key = pyglet.window.key
        if symbol == key.ESCAPE and not (modifiers & ~(key.MOD_NUMLOCK |
                                                       key.MOD_CAPSLOCK |
                                                       key.MOD_SCROLLLOCK)):
            self.dispatch_event('on_close')

        if symbol == key.F5:
            self.start_run = True

        if symbol == key.P:
            self.paused = ~self.paused


class Canvas():

    def __init__(self, size: Tuple[int, int] = (640, 480),
                 origin: Tuple[int, int] = (100, 100)):
        maxsize = (get_screens()[0].width, get_screens()[0].height)
        if size == 'full':
            self.size = maxsize
        else:
            self.size = size

        # check whether origin is outside of window
        # outside = [o<0 for o in origin]
        # self.origin = [o*(1-c) for o,c in zip(origin, outside)]
        self.origin = origin

        self.start_width = size[0]
        self.start_height = size[1]
        self._create_window()

    @property
    def paused(self):
        return self.window.paused

    @property
    def start_run(self):
        return self.window.start_run

    @start_run.setter
    def start_run(self, value: bool):
        self.window.start_run = value

    def estimate_fps(self):
        pyglet.clock.tick()
        for i in range(0, 100, 1):
            self.window.flip()
            pyglet.clock.tick()
        return pyglet.clock.get_fps()

    def _create_window(self):
        self.window = ExperimentalWindow(visible=False,
                                         vsync=True,
                                         width=self.start_width,
                                         height=self.start_height,
                                         resizable=True,
                                         caption='Experimental Framework')

        self.window.set_location(*self.origin)
        self.window.dispatch_events()
        self.window.has_exit = False

    def flip(self):
        "flip the backbuffer to front  and clear the old frontbuffer"
        try:
            self.window.switch_to()
            self.window.dispatch_events()
            self.window.dispatch_event('on_draw')
            self.window.flip()  # flip front to backbuffer
            self.window.clear()  # clear the current backbuffer: was the old backbuffer
        except AttributeError as e:
            if not hasattr(self, 'window') or self.window.has_exit:
                print("Window was closed during flipping")
                return
            else:
                print(__name__, e)
                raise Exception('Window was closed')

    def open(self):
        if not hasattr(self, 'window') or self.window.has_exit:
            self._create_window()
        self.window.set_visible(True)
        self.window.switch_to()
        self.window.dispatch_events()
        self.window.dispatch_event('on_draw')
        self.clear()

    def close(self):
        if hasattr(self, 'window') and not self.window.has_exit:
            self.window.on_close()

    def clear(self):
        "clear both buffers and show a black screen"
        self.flip()
        self.flip()

    def show(self, visual):
        "after having rendered and drawn into the backbuffer, show this"
        try:
            self.window.switch_to()
            for v in visual:
                if v is not None:
                    v.draw(canvas=self)
        # except TypeError: #visual is not iterable
        #    visual.draw(canvas=self)
        except AttributeError:  # window was closed
            print("Window was closed")
            return
        self.flip()

    def set_fullscreen(self):
        self.window.set_fullscreen(fullscreen=True)
        self.flip()

    def set_windowed(self):
        self.window.set_fullscreen(fullscreen=False)
        self.flip()

    def get_width(self):
        return self.window.width

    def get_height(self):
        return self.window.height

    def get_diag(self):
        return (self.height * self.width)**0.5

    width = property(get_width)
    height = property(get_height)
    diag = property(get_diag)

    @property
    def available(self):
        return (hasattr(self, "window") and not self.window.has_exit and self.window.visible)



# -*- coding: utf-8 -*-
"""
Man-in-the-middle
-----------------

Allows to to manage multiple experiments publishing through one LSL-Outlet and
to create the Outlet independently from starting the experiments to improve
detectability and stability for subscribers (e.g. LabRecorder).

Functions and Classes
.....................
"""

import pylsl
import threading
import queue
import time
import socket
import json
from reiz._marker.client import available
import pkg_resources
version = pkg_resources.get_distribution("reiz").version
# %%


class _Outlet():
    "LSL based marker outlet as a singleton, to prevent name-stealing"
    instance = dict()
    @classmethod
    def get(cls, name='reiz-marker'):
        import socket
        import weakref
        source_id = '_at_'.join((name, socket.gethostname()))
        if cls.instance.get(name, None) is None:
            info = pylsl.StreamInfo(name,
                                    type='Markers',
                                    channel_count=1,
                                    nominal_srate=0,
                                    channel_format='string', source_id=source_id)

            info.desc().append_child_value("version", version)
            print(info.as_xml())
            outlet = pylsl.StreamOutlet(info)
            cls.instance[name] = weakref.ref(outlet)
        else:
            outlet = cls.instance[name]()
        return outlet


class _MarkerStreamer(threading.Thread):

    def __init__(self, name: str = None):
        threading.Thread.__init__(self)
        self.queue = queue.Queue(maxsize=0)  # indefinite size
        self.is_running = threading.Event()
        self.name = name

    def push(self, marker: str = '', tstamp: float = None):
        if marker == '':
            return
        if tstamp is None:
            tstamp = pylsl.local_clock()

        self.queue.put_nowait((marker, tstamp))

    def stop(self):
        self.queue.join()
        self.is_running.clear()

    def run(self):
        self.outlet = _Outlet.get(name=self.name)
        self.is_running.set()
        while self.is_running.is_set():
            try:
                marker, tstamp = self.queue.get(block=False)
                self.outlet.push_sample([marker], tstamp)
                self.queue.task_done()
                print(
                    f'Pushed {marker} from {tstamp} at {pylsl.local_clock()}')
            except queue.Empty:
                time.sleep(0.001)
        print(f"Shutting down MarkerStreamer: {self.name}")
# %%


def myip() -> str:
    """returns a string with the computers default IP address
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    ip = s.getsockname()[0]
    s.close()
    return ip


# -----------------------------------------------------------------------------
def _read_msg(client):
    'receive byte for byte to read the header telling the message length'
    # parse the message until it is a valid json
    msg = bytearray(b' ')
    while True:
        try:
            prt = client.recv(1)
            msg += prt
            # because the first byte is b' '
            marker, tstamp = json.loads(msg.decode('ascii'))
            print(f'Received {marker} for {tstamp} at {pylsl.local_clock()}')
            return marker, tstamp
        except json.decoder.JSONDecodeError:
            pass
        except Exception as e:
            print(e)
            break

    return ('', None)


class Server(threading.Thread):
    """Main class to manage the LSL-MarkerStream as man-in-the-middle

    when started, it automatically checks whether there is already a MarkerServer
    running at that port. If this is the case, it returns and lets the old one
    keep control. This ensures that subscribers to the old MarkerServer
    don't experience any hiccups.
    """

    def __init__(self, port: int = 7654, name='reiz-marker',
                 timeout=.05, verbose=True):
        threading.Thread.__init__(self)
        self.host = "127.0.0.1"
        self.port = port
        self.name = name
        self.is_running = threading.Event()
        self.singleton = threading.Event()
        self.verbose = verbose

    def stop(self):
        'stop the server'
        self.is_running.clear()

    def run(self):
        """wait for clients to connect and send messages.

        This is a Thread, so start with :meth:`server.start` """

        # we check whether there is already an instance running, and if so
        # let it keep control by returning
        if available(self.port):
            self.singleton.clear()
            if self.verbose:
                print("Server already running on that port")
            self.is_running.set()
            return
        else:
            self.singleton.set()
            if self.verbose:
                print("This server is the original instance")

        # create the MarkerStreamer, i.e. the LSL-Server that distributes the strings received from the Listener
        markerstreamer = _MarkerStreamer(name=self.name)
        markerstreamer.start()
        # create the ListenerServer, i.e. the TCP/IP Server that waits for messages for forwarding them to the MarkerStreamer
        listener = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        listener.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        listener.settimeout(1)
        listener.bind((self.host, self.port))
        listener.listen(1)
        if self.verbose:
            print('Server mediating an LSL Outlet opened at {0}:{1}'.format(
                self.host, self.port))
        self.is_running.set()
        while self.is_running.is_set():
            try:
                client, address = listener.accept()
                try:
                    marker, tstamp = _read_msg(client)
                    if marker.lower() == "ping":  # connection was only pinged
                        print("Received ping from", address)
                    elif marker.lower() == "poison-pill":
                        print("Swallowing poison pill")
                        self.is_running.clear()
                        break
                    else:
                        markerstreamer.push(marker, tstamp)
                except socket.timeout:
                    print('Client from {address} timed out')
                finally:
                    client.shutdown(2)
                    client.close()
            except socket.timeout:
                pass

        print(f"Shutting down MarkerServer: {self.name}")
        markerstreamer.stop()



# -*- coding: utf-8 -*-
"""
Pyglet based primitive geometric shapes

    
"""
import pyglet            
from pyglet import gl
from pyglet.text import Label

# %% Primitives
'''
The visual primitives have been adapted from cocos2d_. 

   .. _cocos2d https://github.com/los-cocos/cocos

# cocos2d
# Copyright (c) 2008-2012 Daniel Moisset, Ricardo Quesada, Rayentray Tappa,
# Lucio Torre
# Copyright (c) 2009-2017  Richard Jones, Claudio Canepa
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#   * Neither the name of cocos2d nor the names of its
#     contributors may be used to endorse or promote products
#     derived from this software without specific prior written
#     permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
'''
class Base(object):
    """
    Basic attributes of all drawing primitives
    """
    def __init__(self, x, y, z=0, color=(0.,0.,0.,1.), stroke=0, rotation=0):
        try :
            self.rect
        except AttributeError:
            self.rect = Rect(x,y,1,1) # this inits x,y and loc as well
        self.visible = 1 #
        self.z = z 
        self.rotation = rotation
        self.stroke = stroke
        self.color = color
        self.q = gl.gluNewQuadric()
    def setLoc(self, p) : self.rect.loc = p 
    def getLoc(self) : return self.rect.loc
    def setX(self, x) : self.rect.x = x 
    def getX(self) : return self.rect.x 
    def setY(self, y) : self.rect.y = y 
    def getY(self) : return self.rect.y
    loc = property(getLoc, setLoc)
    x = property(getX, setX)
    y = property(getY, setY)
    def setWidth(self,w) : self.rect.width = w
    def getWidth(self) : return self.rect.width
    def setHeight(self, h) : self.rect.height = h
    def getHeight(self) : return self.rect.height
    width = property(getWidth, setWidth)
    height = property(getHeight, setHeight)
        
    def draw(self):
        self.render()
    
class Pixel(Base):
    """ A pixel at a given x,y,z position and color.
        Pixel(x=12, y=100, z=900, color=(1,0,0,0.5))
    """
    def render(self):
        """
            Draws a pixel at a given x and y with given color .
            Color = 3 or 4 arg tuple. RGB values from 0 to 1 being 1 max value (1, 1, 1) would be white
        """
        gl.glColor4f(*self.color)
##        glDisable(GL_TEXTURE_2D) # disable in case it was on
            
        gl.glPushMatrix() # remember previous matrix state before translating, rotating
        gl.glTranslatef(self.x, self.y, -self.z) # translate to point to draw

        gl.glBegin(gl.GL_POINTS) # draw point
        gl.glVertex3f(0.0, 0.0, 0.0)
        gl.glEnd()

        gl.glPopMatrix() # back to previous matrix state

    def intersects(self, x,y):
        if x==self.x and y==self.y : return True


class Circle(Base):
    """ Circle class
        Circle(x=20, y=100, z=1, width=300, color=(1,1,0,0.3), stroke=5, rotation=0, style=GLU_FILL)
        style choices are : GLU_LINE, GLU_FILL, GLU_SILHOUETTE, GLU_POINT
    """
    def __init__(self, x=10, y=10, z=0, width=2, color=(0,0,0,1), stroke=0, rotation=(0,0,0,1), style=gl.GLU_FILL):
        self.radius = width*0.5
        self.rect = Rect(x, y, width, width)
        self.style = style
        self.circleresolution = 60
        Base.__init__(self, x, y, z, color, stroke, rotation)
        
    def setWidth(self, w):
        self.radius = w*0.5
        self.rect.width = w
    width = property(Base.getWidth, setWidth)
        
    def render(self):
        """ Draw Circle
            x, y, z, width in pixel, rotation, color and line width in px
            style choices are : GLU_LINE, GLU_FILL, GLU_SILHOUETTE, GLU_POINT
            TO DO : textured circles
        """
        
        gl.glColor4f(*self.color)
        gl.glPushMatrix()

        gl.glTranslatef(self.x, self.y, -self.z)
        gl.glRotatef(*self.rotation) #angle, bool per axis (x,y,z)


        if self.radius < 1 : self.radius = 1

        if self.stroke :
            inner = self.radius - self.stroke # outline width
            if inner < 0: inner=0
        else :
             inner = 0 # filled
        
        gl.gluQuadricDrawStyle(self.q, self.style)

        gl.gluDisk(self.q, inner, self.radius, self.circleresolution, 1) # gluDisk(quad, inner, outer, slices, loops)
            
        gl.glPopMatrix()



class Arc(Base):
    """ Arc class
        Arc(x=10, y=10, z=0, radius=1, start=0, sweep=1, color=(0,0,0,1), stroke=0, rotation=0.0, style=GLU_FILL)
        style choices are : GLU_LINE, GLU_FILL, GLU_SILHOUETTE, GLU_POINT
    """
    def __init__(self, x=10, y=10, z=0, radius=1, start=0, sweep=1, color=(0,0,0,1), stroke=0,
                 rotation=0.0, style=gl.GLU_FILL):

        Base.__init__(self, x,y,z,color, stroke, rotation)
        self.radius = radius
        self.start = start
        self.sweep = sweep
        self.style = style
        self.circleresolution = 60
        
    def render(self):
        """
        Draws Arc
        """
        gl.glColor4f(*self.color)
        gl.glPushMatrix()
        gl.glTranslatef(self.x, self.y, -self.z)
        gl.glRotatef(self.rotation, 0, 0, 0.1)

        if self.stroke : 
            inner = self.radius - self.stroke
            if inner < 0: inner=0
        else :
            inner = 0 # full, no inner
        self.start -= 180
        
        gl.gluQuadricDrawStyle(self.q, self.style)
        
        gl.gluPartialDisk(self.q, inner, self.radius, self.circleresolution, 1, self.start, self.sweep)
        
        gl.glPopMatrix()

#%%

class Polygon(Base):
    def __init__(self, v, z=0, color=(0,0,0,1), stroke=0, rotation=0.0, style=0):
        """ polygon class
            Polygon(vertexarray=[(0, 0), (29, 100), (30, 200)], z=100, color=(0,0.3,0.1,1), stroke=0, rotation=23)
            overwrites few methods from superclass as polygons are more complex, needs to update everyvertex.
        """
        self.v = v
        l, t, r, b = calcPolygonRect(v) # get the bounding rect
        self.rect = Rect(l+(r-l)*0.5, t+(b-t)*0.5, r-l, b-t)
        self.v2 = [(i[0] - self.rect.x, i[1] - self.rect.y) for i in v] #relative polygon

        self.style = style
        
        Base.__init__(self, self.rect.x, self.rect.y, z,color,stroke,rotation)

    def updateV(self):
        self.v = [(self.rect.x + n[0], self.rect.y + n[1]) for n in self.v2]

    def setLoc(self, p):
        self.rect.loc = p ; self.updateV()
    def setX(self, x):
        self.rect.x = x ; self.updateV()
    def setY(self, y):
        self.rect.y = y; self.updateV()
    x = property(Base.getX, setX)
    y = property(Base.getY, setY)
    loc = property(Base.getLoc, setLoc)
    
    def render(self):
        """ Draw Polygon
            v is an array with tuple points like [(x, y), (x2, y2), (x3, y3)]
            min vertex number to draw a polygon is 3
            stroke=0 to fil with color the inside of the shape or stroke=N just to draw N-px thick outline.
            Note. It doesnt work with non convex polygons, need to implement tesselation yet
        """
        l,t,r,b = calcPolygonRect(self.v)
        x,y = calcRectCenter(l,t,r,b)
        self.drawVertex(x, y, self.z, [(i[0] - x, i[1] - y) for i in self.v], self.color, self.stroke, self.rotation, self.style)

    
    def drawVertex(self, x, y,  z=0, v=(), color=(0,0,0,1), stroke=0, rotation=0.0,   style=0):
        
        gl.glColor4f(*self.color)    
        gl.glPushMatrix()

        gl.glTranslatef(x, y, -z)
        gl.glRotatef(self.rotation, 0, 0, 0.1)

        if self.style :
            gl.glEnable(gl.GL_LINE_STIPPLE)
            gl.glLineStipple(1, style)
##        else :
##            glDisable(GL_LINE_STIPPLE)
##            0xF0F0 # dashed line
##            0xF00F # long dashed line
##            0x8888 # dotted lines
##        glRect(x1,y,1,x1,x2)
##        glRectiv(v1,v2) # oposite vertex of rectangle
        # -- start drawing
        if self.stroke : # outlined polygon
            gl.glLineWidth(self.stroke)
            gl.glBegin(gl.GL_LINE_LOOP)
        else: # filled polygon
            if   len(v) == 4 : gl.glBegin(gl.GL_QUADS)
            elif len(v)  > 4 : gl.glBegin(gl.GL_POLYGON)
            else :             gl.glBegin(gl.GL_TRIANGLES) # which type of polygon are we drawing?

        for p in v:
            gl.glVertex3f(p[0], p[1],0)  # draw each vertex

        gl.glEnd()
        # -- end drawing
        
        if self.style : gl.glDisable(gl.GL_LINE_STIPPLE)
        
        gl.glPopMatrix()


class LineRel(Base):
    def __init__(self, x,y, a=(0,0), b=(0,0), z=0, color=(0,0,0,1), stroke=0, rotation=0.0, style=0):
        """ Draws a basic line given the begining and end point (tuples), color (tuple) and stroke
            (thickness of line)
            Line( x,y, a=(1,1), b=(100,100), z=0, color=(0.2,0,0,1), stroke=10, rotation=45)
        """
        w = (b[0] - a[0]) 
        h = (b[1] - a[1]) 
        x = abs(a[0] + w*0.5)
        y = abs(a[1] + h*0.5)
        self.a2 = abs(a[0]) - x, abs(a[1]) - y
        self.b2 = abs(b[0]) - x, abs(b[1]) - y
        self.a = x - w*0.5, y - w*0.5
        self.b = x + w*0.5, y + w*0.5
        self.rect = Rect(x, y, w, h)
        self.style = style
        Base.__init__(self, x, y, z,color,stroke,rotation)

    def render(self):
        """
        Draws Line
        """
        p1 = self.a2
        p2 = self.b2
        gl.glColor4f(*self.color)
        color  = (gl.GLfloat *4)(*self.color)
               
        gl.glPushMatrix()

        gl.glTranslatef(self.x, self.y, -self.z) # translate to GL loc ppint
        gl.glRotatef(self.rotation, 0, 0, 0.1)

        if self.style :
            gl.glEnable(gl.GL_LINE_STIPPLE)
            gl.glLineStipple(1, self.style)
##        else :
##            glDisable(GL_LINE_STIPPLE)
            
        if self.stroke <= 0:
            self.stroke = 1
        gl.glLineWidth(self.stroke)

        gl.glBegin(gl.GL_LINES)
        gl.glVertex2i(int(p1[0]), int(p1[1])) # draw pixel points
        gl.glVertex2i(int(p2[0]), int(p2[1]))
        gl.glEnd()

        if self.style :
            gl.glDisable(gl.GL_LINE_STIPPLE)
        
        gl.glPopMatrix()


    def updateAB(self):
        self.a = self.x + self.a[0], self.y + self.a[0]
        self.b = self.x + self.b[0], self.y + self.b[0]

    def setLoc(self, p):
        self.rect.loc = p ; self.updateAB()
    def setX(self, x):
        self.rect.x = x ; self.updateAB()
    def setY(self, y):
        self.rect.y = y; self.updateAB()
    x = property(Base.getX, setX)
    y = property(Base.getY, setY)
    loc = property(Base.getLoc, setLoc)


class Line(LineRel):
    def __init__(self, a=(0,0), b=(0,0), z=0, color=(0,0,0,1), stroke=0, rotation=0.0, style=0):
        """ Draws a basic line given the begining and end point (tuples), color (tuple) and stroke
            (thickness of line)
            Line( a=(1,1), b=(100,100), z=20, color=(0.2,0,0,1), stroke=10, rotation=45)
        """
        w = (b[0] - a[0]) 
        h = (b[1] - a[1]) 
        x = abs(a[0] + w*0.5) # abs x,y
        y = abs(a[1] + h*0.5)
        a = x-w*0.5, y-h*0.5 # relative a,b
        b = x+w*0.5, y+h*0.5
        LineRel.__init__(self, x, y, a, b, z, color, stroke, rotation, style)



# intersectable shapes (draggable)

class Rect(object):
    def __init__(self, x=0,y=0,w=0,h=0):
        """ rect(self, x=0,y=0,w=0,h=0)
            x,y,loc, width, height
            left,top,right,bottom
            quad ->
            --------------------
            topleft = quad[0]
            topright = quad[1]
            bottomright = quad[2]
            bottomleft = quad[3]
        """
        self.rect = x,y,w,h

    def setRect(self, r):
        self.__x = r[0]
        self.__y = r[1]
        self.__width = r[2]
        self.__height = r[3]
        w = r[2]*0.5 ; h = r[3]*0.5
        self.__rect = r[0]-w, r[1]-h, r[0]+w, r[1]+h # l t r b
    def getRect(self):
        return self.__rect
    rect = property(getRect, setRect)

    def setQuad(self, q): # [ q[0][0], q[0][1], q[1][0], q[2][1] ] # l t r b
        self.rect = q[0][0]+(q[1][0]-q[0][0])*0.5, q[0][1]+(q[2][1]-q[0][1])*0.5, q[1][0]-q[0][0], q[2][1]-q[0][1] 
    def getQuad(self):
        return [(self.rect[0], self.rect[1]),(self.rect[2], self.rect[1]),(self.rect[2], self.rect[3]),(self.rect[0], self.rect[3])] # tl tr br bl
    quad = property(getQuad, setQuad)

    def setX(self, x):
        self.rect = x, self.y, self.width, self.height
    def getX(self) : return self.__x
    x = property(getX, setX)

    def setY(self, y):
        self.rect = self.x, y, self.width, self.height
    def getY(self) : return self.__y
    y = property(getY, setY)

    def setLoc(self, p):
        self.rect = p[0], p[1], self.width, self.height
    def getLoc(self) : return self.__x, self.__y # self.x, self.y
    loc = property(getLoc, setLoc)

    def setWidth(self, w):
        self.rect = self.x, self.y, w, self.height
    def getWidth(self): return self.__width
    width = property(getWidth, setWidth)
    
    def setHeight(self, h):
        self.rect = self.x, self.y, self.width, h
    def getHeight(self): return self.__height
    height = property(getHeight, setHeight)

    def setLeft(self,x):
        self.rect = x+self.width*0.5, self.y, self.width, self.height
    def getLeft(self): return self.rect[0]
    left = property(getLeft, setLeft)
    
    def setTop(self,y):
        self.rect = self.x, y+self.width*0.5, self.width, self.height
    def getTop(self): return self.rect[1]
    top = property(getTop, setTop)
    
    def setRight(self,x):
        self.rect = x-self.width*0.5, self.y, self.width, self.height
    def getRight(self): return self.rect[2]
    right = property(getRight, setRight)
    
    def setBottom(self,x):
        self.rect = self.x, self.y-self.width*0.5, self.width, self.height
    def getBottom(self): return self.rect[3]
    bottom = property(getBottom, setBottom)


        
#%%
def calcPolygonRect(pointArray):
    """ receives a point list and returns the rect that contains them as a tupple -> tuple left, top, right, bottom
    """
    # init to ridiculously big values. not very elegant or eficient
    l, t, r, b = 10000000, 10000000, -10000000, -10000000
##    l = pointArray[0]
##    t = pointArray[1]
##    r = l
##    b = t

    for n in pointArray: # calc bounding rectangle rect
        if n[0] < l : l = n[0]
        if n[0] > r : r = n[0]
        if n[1] < t : t = n[1]
        if n[1] > b : b = n[1]

    return l, t, r, b


def calcRectCenter(l,t,r,b):#,v=()):
    """ returns rect center point -> x,y
        calcRectCenter(l,t,r,b)
    """
##    if len(v) : l,t,r,b = v[0],v[1],v[2],v[3]
    return l+((r-l)*0.5), t+((b-t)*0.5)





from pytest import fixture
import reiz


@fixture
def rmarker():
    # set up
    assert reiz.marker.available() is False
    reiz.clock.reset()
    reiz.marker.start()
    assert reiz.clock.now() < 10  # create within 10 secs
    # interrupt
    yield True
    # tear down
    reiz.clock.reset()
    reiz.marker.stop()
    assert reiz.clock.now() < 10  # kill within 10 secs


def test_sanitization(rmarker, capsys):
    assert rmarker
    out, err = capsys.readouterr()
    reiz.marker.push("ä ö ü")
    out, err = capsys.readouterr()
    assert "Sending ae_oe_ue " in out
    out, err = capsys.readouterr()
    reiz.marker.push("ping")
    out, err = capsys.readouterr()
    assert 'Sending {\"msg": "ping"}' in out
    out, err = capsys.readouterr()
    reiz.marker.push("poison-pill")
    out, err = capsys.readouterr()
    assert 'Sending {\"msg": "poison-pill"}' in out

    c = reiz.Canvas((10, 10))

    cue = reiz.Cue(canvas=c,
                   audiostim=reiz.audio.library.beep, visualstim=reiz.visual.library.go,
                   markerstr="test")

    out, err = capsys.readouterr()
    cue.show()
    out, err = capsys.readouterr()
    assert 'Sending test at' in out



import setuptools
from distutils.core import setup
# read the contents of README.md
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='Reiz',
    version="v0.3.3.3",
    description='A Python toolbox for visual and auditory stimulation based on pyglet and pylsl.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Robert Guggenberger',
    author_email='robert.guggenberger@uni-tuebingen.de',
    url='https://github.com/pyreiz/pyreiz',
    download_url='https://github.com/pyreiz/pyreiz.git',
    license='MIT',
    include_package_data=True,
    package_data={'reiz': ['data/*.*']},
    packages=setuptools.find_packages(),
    install_requires=[
        'pyglet >= 1.4.7',
        'pylsl >= 1.13',
    ],
    extras_require={'tts': ['pyttsx3 >= 2.7']},
    entry_points={"console_scripts":
                  ["reiz-marker=reiz._marker.__main__:main"]},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Multimedia :: Graphics :: Presentation',
        'Topic :: Multimedia :: Sound/Audio :: Players',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces',
        'Topic :: Scientific/Engineering :: Medical Science Apps.'
    ]
)



import reiz
from pytest import fixture


@fixture
def stimuli():
    c = reiz.Canvas((10, 10))
    cue = reiz.Cue(canvas=c,
                   audiostim=reiz.audio.library.beep,  visualstim=reiz.visual.library.go)
    N = reiz.Cue(canvas=c,
                 audiostim=None,
                 visualstim=None)
    return c, cue, N


def test_cue(stimuli):
    c, cue, N = stimuli
    reiz.audio.library.beep.volume = 0
    c.open()
    cue.show(0.5)
    N.show(canvas=c)
    N.show()
    cue.show(None)
    c.close()


def test_collection(stimuli):
    c, cue, N = stimuli
    lib = reiz.cue.collect(**{"N": N, "C": cue})
    assert "N" in lib.__dict__
    assert "C" in lib.__dict__
    assert lib.N is N
